/*
    Project: TCGINA
    Version: V1.26
    File   : kfidecrypt.c

    Copyright (c) 2005-2009 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "kfipriv.h"

#ifdef _MSC_VER
#  pragma warning( push )
#  pragma warning( disable : 4701 )
#endif

int KFI_Decrypt(KFI_DATA *pdDest, const KFI_DATA *pdSrc,
  const KFI_ENCRYPTION_DATA *ped)
{
    int               iCipher1;
    int               iCipher2;
    int               iError;
    BYTE              *pbDest;
    BYTE              *pbDestStart;
    BYTE              *pbDestEnd;
    KFI_DATA *pd;
    struct {
        BYTE          rgbState[KFI_CIPHER_BLOCK_SIZE];
        BYTE          rgbPrevCT[KFI_CIPHER_BLOCK_SIZE];
        symmetric_key skKey1;
        symmetric_key skKey2;
    } *tmp;

    if (ped->fValid) {
        if (register_cipher(&KFI_CIPHER1_DESC) == -1) {
            return CRYPT_INVALID_CIPHER;
        }
        if ((iCipher1 = find_cipher(KFI_CIPHER1_NAME)) == -1) {
            return CRYPT_INVALID_CIPHER;
        }
    } else {
        /* Verify signature and checksum */
        if (pdSrc->dwSignature == KFI_SIGNATURE) {
            if (calc_crc32(((unsigned char*)pdSrc) + KFI_MESSAGE_OFFSET,
              KFI_MESSAGE_SIZE) == pdSrc->dwCRC) {
                iError = CRYPT_OK;
                goto already_decrypted;
            }
        }
    }
    if (register_cipher(&KFI_CIPHER2_DESC) == -1) {
        return CRYPT_INVALID_CIPHER;
    }
    if ((iCipher2 = find_cipher(KFI_CIPHER2_NAME)) == -1) {
        return CRYPT_INVALID_CIPHER;
    }

    if (!sec_alloc(tmp)) {
        return CRYPT_MEM;
    }
    if (pdDest == pdSrc) {
        if ((pd = sec_malloc(sizeof(KFI_DATA))) == NULL) {
            sec_free(tmp);
            return CRYPT_MEM;
        }
    } else {
        pd = pdDest;
    }
    memcpy(pd, pdSrc, sizeof(KFI_DATA));

    /* Decrypt data */
    if (ped->fValid) {
        if ((iError = cipher_descriptor[iCipher1].setup(ped->rgbKey1,
          sizeof(ped->rgbKey1), 0, &tmp->skKey1)) != CRYPT_OK) {
            goto error;
        }
        if ((iError = cipher_descriptor[iCipher2].setup(ped->rgbKey2,
          sizeof(ped->rgbKey2), 0, &tmp->skKey2)) != CRYPT_OK) {
            goto error;
        }
        pbDestStart = ((unsigned char *)pd) + KFI_DATA_OFFSET;
        pbDestEnd = pbDestStart + KFI_DATA_SIZE;

        pbDest = pbDestStart;
        while (pbDest != pbDestEnd) {
            KFI_BlockXor(pbDest, ped->rgbWhiteningKey);
            pbDest += KFI_CIPHER_BLOCK_SIZE;
        }
        pbDest = pbDestStart;
        while (pbDest != pbDestEnd) {
            memcpy(tmp->rgbState, pbDest, KFI_CIPHER_BLOCK_SIZE);
            cipher_descriptor[iCipher2].ecb_decrypt(pbDest, pbDest,
              &tmp->skKey2);
            cipher_descriptor[iCipher1].ecb_decrypt(pbDest, pbDest,
              &tmp->skKey1);
            KFI_BlockXor(pbDest, (pbDest != pbDestStart) ?
              tmp->rgbPrevCT : ped->rgbIV);
            memcpy(tmp->rgbPrevCT, tmp->rgbState, KFI_CIPHER_BLOCK_SIZE);
            pbDest += KFI_CIPHER_BLOCK_SIZE;
        }
    } else {
        if ((iError = cipher_descriptor[iCipher2].setup(pd->rgbSalt,
          KFI_CIPHER_KEY_SIZE, 0, &tmp->skKey2)) != CRYPT_OK) {
            goto error;
        }
        pbDestStart = ((unsigned char *)pd) + KFI_DATA_OFFSET;
        pbDestEnd = pbDestStart + KFI_DATA_SIZE;
        pbDest = pbDestStart;
        while (pbDest != pbDestEnd) {
            memcpy(tmp->rgbState, pbDest, KFI_CIPHER_BLOCK_SIZE);
            cipher_descriptor[iCipher2].ecb_decrypt(pbDest, pbDest,
              &tmp->skKey2);
            KFI_BlockXor(pbDest, (pbDest != pbDestStart) ?
              tmp->rgbPrevCT : (pd->rgbSalt+KFI_CIPHER_KEY_SIZE));
            memcpy(tmp->rgbPrevCT, tmp->rgbState, KFI_CIPHER_BLOCK_SIZE);
            pbDest += KFI_CIPHER_BLOCK_SIZE;
        }
    }

    /* Verify signature and checksum */
    if (pd->dwSignature != KFI_SIGNATURE) {
        iError = CRYPT_ERROR;
        goto error;
    }
    if (calc_crc32(((unsigned char*)pd) + KFI_MESSAGE_OFFSET,
      KFI_MESSAGE_SIZE) != pd->dwCRC) {
        iError = CRYPT_ERROR;
        goto error;
    }

already_decrypted:
    if (pd->wType != 0 || pd->wVersion != 0) {
        iError = CRYPT_ERROR;
        goto error;
    }
    pd->wszHostName[KFI_MAX_HOST_NAME-1] = L'\0';
    pd->wszDescription[KFI_MAX_DESCRIPTION-1] = L'\0';
    pd->aszPassword[KFI_MAX_VOLUME_PASSWORD-1] = '\0';
    if (pd->wPasswordSize == 0 || pd->wPasswordSize > KFI_MAX_PASSWORD_LEN) {
        pd->wPasswordSize = (WORD) strlen(pd->aszPassword);
    }
    if (pd->chDriveLetter) {
        BYTE bDrive = (BYTE)pd->chDriveLetter;
        if (bDrive >= 'A' && bDrive <= 'Z') {
        } else if (bDrive < 27) {
            pd->chDriveLetter = bDrive + ('A' - 1);
        } else if (bDrive >= 'a' && bDrive <= 'z') {
            pd->chDriveLetter = bDrive - ('a' - 'A');
        } else {
            pd->chDriveLetter = 0;
        }
    }

    if (pdDest == pdSrc) {
        memcpy(pdDest, pd, sizeof(KFI_DATA));
    }

error:
    if (pdDest == pdSrc) {
        sec_free(pd);
    }
    sec_free(tmp);

    return iError;
}

#ifdef _MSC_VER
#  pragma warning( pop )
#endif
